//
// This sample application was written to work with the 'BareBones' variant of the PixeLINK
// API.  That is a pre-production version of the PixeLINK API that had much reduced functionality.
// It is still included with the full API variant for historical reasons, as a kind of 'bridge'
// for those customers that are transitioning from using the BareBones variant of the API, to
// the full API.
//
// This appication will work with the full PixeLINK API, but new development should use the much
// more comprehesive features of the full API to interact with the camera.  Those users should have
// a look at the 'getCameraFeature', 'whiteBalance', or 'autoExposure' sample applications to get
// a better idea of how to interact with the API/camera using those features.
//
#include "PixeLINKApi.h"
#include "LinuxUtil.h"

#include <iostream>
#include <stdio.h>
#include <stdexcept>
#include <unistd.h>
#include <vector>

using namespace std;

// PxLPrivateCommands used to read and write vision registers
#define VISION_LWORD_READ  43
#define VISION_LWORD_WRITE 44

// Vision registers
#define EXPOSURE_REG 0x60084


typedef struct _CAMERA_OP
{
    U32 opCode;   // One of VISION_LWORD_READ or VISION_LWORD_WRITE
    U32 reg;      // Register to be read/written
    U32 value;    // register value (read or written)
} CAMERA_OP;

// Returns the current exposure of the camera, in milliseconds.  Returns 0.0 on error
static float getExposure (HANDLE myCamera)
{
    ULONG rc;
    CAMERA_OP op;

    op.opCode = VISION_LWORD_READ;
    op.reg   = EXPOSURE_REG;
    op.value = 0x55555555;  // give it a bogus value.
    rc = PxLPrivateCmd (myCamera, sizeof (CAMERA_OP), (U32*)&op);
    if (!(API_SUCCESS(rc)))
    {
         printf ("        Could not read register 0x%X, rc=0x%X\n", op.reg, rc);
         return 0.0f;
    }

    return (*(float*)&op.value) / 1000.0f;
}

// Sets the exposure of the camera, to the supplied (millisecond) value
static void setExposure (HANDLE myCamera, float newExposure)
{
    ULONG rc;
    CAMERA_OP op;

    newExposure *= 1000.0f;  // convert milliseconds to the microseconds required by the register.

    op.opCode = VISION_LWORD_WRITE;
    op.reg   = EXPOSURE_REG;
    op.value = *(ULONG*)&newExposure; // casting permits float as a U32;

    rc = PxLPrivateCmd (myCamera, sizeof (CAMERA_OP), (U32*)&op);
    if (!(API_SUCCESS(rc)))
    {
         printf ("        Could not write register 0x%X, rc=0x%X\n", op.reg, rc);
    }
}

int main() {
    HWND  previewWin;
    HANDLE myCamera;

    DontWaitForEnter unbufferedKeyboard;  // Declare this for our getchar
    int   keyPressed;
    bool  done = false;
    bool previewing = false;

    ULONG rc;

    rc = PxLInitializeEx(0,&myCamera,0);
    if (!API_SUCCESS(rc))
    {
        printf ("Could not Initialize the camera!  Rc = 0x%X\n", rc);
        return 1;
    }

    printf ("Starting the stream for camera with handle:%p, expsoure:%f ms\n", myCamera, getExposure(myCamera));
    printf ("    q   : to quit\n");
    printf ("    p   : toggles the preview window on/off\n");
    printf ("    g   : grabs a frame (via PxLGetNextFrame)\n");
    printf ("    +/- : changes the exposure by 10%% (up or down)\n");

    // Starting the stream will instantiate the the BufferControl -- so do it before I start the preview
    rc = PxLSetStreamState (myCamera, START_STREAM);
    if (!API_SUCCESS(rc))
    {
        printf ("Could not start the stream! Rc = 0x%X\n", rc);
        PxLUninitialize(myCamera);
        return 1;
    }

    // Just going to declare a very large buffer here
    // One that's large enough for any PixeLINK 4.0 camera
    std::vector<U8> frameBuffer(3000*3000*2);
    FRAME_DESC frameDesc;

    while (!done)
    {
        fflush(stdin);
        if (kbhit())
        {
            keyPressed = getchar();
            switch (keyPressed)
            {
            case 'q':
            case 'Q':
                printf ("\n\n");
                done = true;
                break;
            case 'p':
            case 'P':
                previewing = !previewing;
                rc = PxLSetPreviewState (myCamera, previewing ? START_PREVIEW : STOP_PREVIEW, &previewWin);
                break;
            case 'g':
            case 'G':
                rc = PxLGetNextFrame(myCamera, (U32)frameBuffer.size(), &frameBuffer[0], &frameDesc);
                if (API_SUCCESS(rc))
                {
                    printf ("  FrameTime:%5.3f, FrameNumber:%d\n", frameDesc.fFrameTime, frameDesc.uFrameNumber);
                } else {		
                    printf ("Grabbing a frame returned 0x%X\n", rc);
                }
                break;
            case '+':
            case '-':
                {
                    float exposure = getExposure (myCamera);
                    if (keyPressed == '+')
                        exposure *= 1.1f; // increase by 10%
                    else
                        exposure *= 0.9f; // decrease by 10%
                    setExposure (myCamera, exposure);
                    printf ("Exposure set to %4.3f ms\n", exposure);
                }
                break;
            }
            fflush(stdout);
        }
        usleep (50*1000);  // 50 ms - don't hog all of the CPU
    }

    if (previewing)
        PxLSetPreviewState(myCamera, STOP_PREVIEW, &previewWin);

	PxLSetStreamState (myCamera, STOP_STREAM);

	PxLUninitialize (myCamera);
}

